use core::net::Ipv4Addr;

macro_rules! define_dhcp_options {
    ($($code:expr, $name:ident, $data:ty),+) => {
        enum Code {
            $($name = $code),+
        }
        $(
            pub struct $name(pub $data);

            impl DhcpOption for $name {
                const CODE: u8 = $code;
                type Item = $data;

                fn parse(option: &RawDhcpOption) -> Result<Self::Item, DhcpOptionError> {
                    TryFrom::try_from(option)
                }
            }
        )+
    };
}

#[rustfmt::skip]
define_dhcp_options!(
    1, SubnetMask, Ipv4Addr,
    2, TimeOffset, u32,
    3, RouterOption, Vec<Ipv4Addr>,
    4, TimeServerOption, Vec<Ipv4Addr>,
    5, NameServerOption, Vec<Ipv4Addr>,
    6, DomainNameServerOption, Vec<Ipv4Addr>,
    7, LogServerOption, Vec<Ipv4Addr>,
    8, CookieServerOption, Vec<Ipv4Addr>,
    9, LprServerOption, Vec<Ipv4Addr>,
    10, ImpressServerOption, Vec<Ipv4Addr>,
    11, ResourceLocationServerOption, Vec<Ipv4Addr>,
    12, HostNameOption, String,
    13, BootFileSizeOption, u16,
    14, MeritDumpFile, String,
    15, DomainName, String,
    16, SwapServer, Vec<Ipv4Addr>,
    17, RootPath, String,
    18, ExtensionPath, String,
    19, IPForwardingOption, bool,
    20, NonLocalSourceRoutingOption, bool,
    21, PolicyFilterOption, Vec<(Ipv4Addr, Ipv4Addr)>,
    22, MaximumDatagramReassemblySize, u16,
    23, DefaultIPTimeToLive, u8,
    24, PathMTUAgingTimeoutOption, u32,
    25, PathMTUPlateauTableOption, Vec<u16>,
    26, InterfaceMTUOption, u16,
    27, AllSubnetsAreLocalOption, bool,
    28, BroadcastAddressOption, Ipv4Addr,
    29, PerformMaskDiscoveryOption, bool,
    30, MaskSupplierOption, bool,
    31, PerformRouterDiscoveryOption, bool,
    32, RouterSolicitationAddressOption, Ipv4Addr,
    33, StaticRouteOption, Vec<(Ipv4Addr, Ipv4Addr)>,
    34, TrailerEncapsulationOption, bool,
    35, ARPCacheTimeoutOption, u32,
    36, EthernetEncapsulationOption, bool,
    37, TCPDefaultTTLOption, u8,
    38, TCPKeepaliveIntervalOption, u32,
    39, TCPKeepaliveGarbageOption, bool,
    40, NetworkInformationServiceDomainOption, String,
    41, NetworkInformationServersOption, Vec<Ipv4Addr>,
    42, NetworkTimeProtocolServersOption, Vec<Ipv4Addr>,
    43, VendorSpecificInformation, Vec<u8>,
    44, NetBIOSOverTCPIPNameServerOption, Vec<Ipv4Addr>,
    45, NetBIOSOverTCPIPDatagramDistributionServerOption, Vec<Ipv4Addr>,
    46, NetBIOSOverTCPIPNodeTypeOption, u8, //TODO: Make separate type. RFC2132 section 8.6
    47, NetBIOSOverTCPIPScopeOption, String,
    48, XWindowSystemFontServerOption, Vec<Ipv4Addr>
    
);

impl<'a> TryFrom<&RawDhcpOption<'a>> for Ipv4Addr {
    type Error = DhcpOptionError;

    fn try_from(option: &RawDhcpOption<'a>) -> Result<Self, Self::Error> {
        (option.length == 4)
            .then_some(())
            .ok_or(DhcpOptionError::InvalidLength(option.length, 4))?;
        Ok(Ipv4Addr::new(
            option.data[0],
            option.data[1],
            option.data[2],
            option.data[3],
        ))
    }
}

impl<'a> TryFrom<&RawDhcpOption<'a>> for u16 {
    type Error = DhcpOptionError;

    fn try_from(option: &RawDhcpOption<'a>) -> Result<Self, Self::Error> {
        (option.length == 2)
            .then_some(())
            .ok_or(DhcpOptionError::InvalidLength(option.length, 2))?;
        let data = option.data[..2].try_into().expect("Parser pulled 2 bytes");
        Ok(u16::from_be_bytes(data))
    }
}

impl<'a> TryFrom<&RawDhcpOption<'a>> for Vec<u16> {
    type Error = DhcpOptionError;

    fn try_from(option: &RawDhcpOption<'a>) -> Result<Self, Self::Error> {
        let rem = option.length % 2;
        (rem == 0)
            .then_some(())
            .ok_or(DhcpOptionError::InvalidLength(
                option.length,
                option.length + 2 - rem,
            ))?;
        Ok(option
            .data
            .chunks(2)
            .map(|c| u16::from_be_bytes(c.try_into().expect("validated before")))
            .collect())
    }
}

impl<'a> TryFrom<&RawDhcpOption<'a>> for u8 {
    type Error = DhcpOptionError;

    fn try_from(option: &RawDhcpOption<'a>) -> Result<Self, Self::Error> {
        (option.length == 1)
            .then_some(())
            .ok_or(DhcpOptionError::InvalidLength(option.length, 1))?;
        Ok(option.data[0])
    }
}

impl<'a> TryFrom<&RawDhcpOption<'a>> for u32 {
    type Error = DhcpOptionError;

    fn try_from(option: &RawDhcpOption<'a>) -> Result<Self, Self::Error> {
        (option.length == 4)
            .then_some(())
            .ok_or(DhcpOptionError::InvalidLength(option.length, 2))?;
        let data = option.data[..4].try_into().expect("Parser pulled 4 bytes");
        Ok(u32::from_be_bytes(data))
    }
}

impl<'a> TryFrom<&RawDhcpOption<'a>> for Vec<Ipv4Addr> {
    type Error = DhcpOptionError;

    fn try_from(option: &RawDhcpOption<'a>) -> Result<Self, Self::Error> {
        let rem = option.length % 4;
        (rem == 0)
            .then_some(())
            .ok_or(DhcpOptionError::InvalidLength(
                option.length,
                option.length + 4 - rem,
            ))?;
        Ok(option
            .data
            .chunks(4)
            .map(|c| Ipv4Addr::new(c[0], c[1], c[2], c[3]))
            .collect())
    }
}

impl<'a> TryFrom<&RawDhcpOption<'a>> for Vec<u8> {
    type Error = DhcpOptionError;

    fn try_from(option: &RawDhcpOption<'a>) -> Result<Self, Self::Error> {
        Ok(option.data.to_vec())
    }
}

impl<'a> TryFrom<&RawDhcpOption<'a>> for String {
    type Error = DhcpOptionError;

    fn try_from(option: &RawDhcpOption<'a>) -> Result<Self, Self::Error> {
        let data = core::str::from_utf8(option.data)?;
        Ok(data.to_string())
    }
}

impl<'a> TryFrom<&RawDhcpOption<'a>> for bool {
    type Error = DhcpOptionError;

    fn try_from(option: &RawDhcpOption<'a>) -> Result<Self, Self::Error> {
        Ok(option.data[0] == 1)
    }
}

impl<'a> TryFrom<&RawDhcpOption<'a>> for Vec<(Ipv4Addr, Ipv4Addr)> {
    type Error = DhcpOptionError;

    fn try_from(option: &RawDhcpOption<'a>) -> Result<Self, Self::Error> {
        let rem = option.length % 8;
        (rem == 0)
            .then_some(())
            .ok_or(DhcpOptionError::InvalidLength(
                option.length,
                option.length + 8 - rem,
            ))?;
        Ok(option
            .data
            .chunks(8)
            .map(|c| {
                (
                    Ipv4Addr::new(c[0], c[1], c[2], c[3]),
                    Ipv4Addr::new(c[4], c[5], c[6], c[7]),
                )
            })
            .collect())
    }
}

#[derive(Debug)]
pub struct RawDhcpOption<'a> {
    pub code: u8,
    pub length: u8,
    pub data: &'a [u8],
}

pub trait DhcpOption {
    const CODE: u8;
    type Item;

    fn parse(option: &RawDhcpOption) -> Result<Self::Item, DhcpOptionError>;
}

#[derive(Debug)]
pub enum DhcpOptionError {
    InvalidLength(u8, u8),
    UTF8Error,
}

impl core::fmt::Display for DhcpOptionError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            DhcpOptionError::InvalidLength(got, should) => {
                write!(f, "Invalid length got {got}, should be {should}")
            }
            DhcpOptionError::UTF8Error => write!(f, "Error parsing option data as UTF-8"),
        }
    }
}

impl core::error::Error for DhcpOptionError {}

impl From<core::str::Utf8Error> for DhcpOptionError {
    fn from(value: core::str::Utf8Error) -> Self {
        DhcpOptionError::UTF8Error
    }
}
